@@ -6,11 +6,10 @@ use crate::{
     utils::{Filter, Tokenizer},
 };
 use std::{
-    cell::RefCell,
     collections::{HashMap, HashSet},
     fmt::Debug,
     hash::Hash,
-    rc::Rc,
+    sync::{Arc, RwLock},
 };
 
 use self::score::calculator::{FieldData, TermData};
@@ -85,34 +84,34 @@ pub fn query<T: Eq + Hash + Clone + Debug, M, S: ScoreCalculator<T, M>>(
             let mut visited_documents_for_term: HashSet<T> = HashSet::new();
             for query_term_expanded in expanded_terms {
                 let term_node_option =
-                    find_inverted_index_node(Rc::clone(&index.root), &query_term_expanded);
+                    find_inverted_index_node(Arc::clone(&index.root), &query_term_expanded);
                 if let Some(term_node) = term_node_option {
-                    let mut term_node_borrowed = term_node.borrow_mut();
+                    let mut term_node_borrowed = term_node.write().unwrap();
                     let mut new_first_doc = None;
                     let mut assign_new_first_doc = false;
                     let mut document_frequency = 0;
 
                     if let Some(term_node_option_first_doc) = &mut term_node_borrowed.first_doc {
-                        let mut prev_pointer: Option<Rc<RefCell<DocumentPointer<T>>>> = None;
-                        let mut pointer_option = Some(Rc::clone(&term_node_option_first_doc));
+                        let mut prev_pointer: Option<Arc<RwLock<DocumentPointer<T>>>> = None;
+                        let mut pointer_option = Some(Arc::clone(&term_node_option_first_doc));
                         while let Some(pointer) = pointer_option {
                             if removed.is_some() // Cleanup old removed documents while searching. If vaccume after delete, this will have not effect
                                 && removed
                                     .unwrap()
-                                    .contains(&pointer.borrow().details.borrow().key)
+                                    .contains(&pointer.read().unwrap().details.read().unwrap().key)
                             {
                                 if let Some(pp) = &prev_pointer {
-                                    pp.borrow_mut().next = pointer.borrow().next.clone();
+                                    pp.write().unwrap().next = pointer.read().unwrap().next.clone();
                                 } else {
-                                    new_first_doc = (&pointer.borrow().next).clone();
+                                    new_first_doc = (&pointer.read().unwrap().next).clone();
                                     assign_new_first_doc = true;
                                     //  term_node_borrowed.first_doc = (&pointer.borrow().next).clone();
                                 }
                             } else {
-                                prev_pointer = Some(Rc::clone(&pointer));
+                                prev_pointer = Some(Arc::clone(&pointer));
                                 document_frequency += 1;
                             }
-                            pointer_option = pointer.borrow().next.clone();
+                            pointer_option = pointer.read().unwrap().next.clone();
                         }
                     }
 
@@ -133,18 +132,19 @@ pub fn query<T: Eq + Hash + Clone + Debug, M, S: ScoreCalculator<T, M>>(
                                 docs,
                             );
 
-                            let mut pointer = Some(Rc::clone(&term_node_option_first_doc));
+                            let mut pointer = Some(Arc::clone(&term_node_option_first_doc));
                             while let Some(p) = pointer {
-                                let pointer_borrowed = p.borrow();
-                                let field_lengths = &pointer_borrowed.details.borrow().field_length;
+                                let pointer_borrowed = p.read().unwrap();
+                                let field_lengths =
+                                    &pointer_borrowed.details.read().unwrap().field_length;
                                 if removed.is_none()
                                     || !removed
                                         .unwrap()
-                                        .contains(&pointer_borrowed.details.borrow().key)
+                                        .contains(&pointer_borrowed.details.read().unwrap().key)
                                 {
                                     let score = &score_calculator.score(
                                         pre_calculations.as_ref(),
-                                        p.borrow(),
+                                        p.read().unwrap(),
                                         &FieldData {
                                             field_lengths,
                                             fields_boost,
@@ -153,7 +153,7 @@ pub fn query<T: Eq + Hash + Clone + Debug, M, S: ScoreCalculator<T, M>>(
                                         &term_expansion_data,
                                     );
                                     if let Some(s) = score {
-                                        let key = &pointer_borrowed.details.borrow().key;
+                                        let key = &pointer_borrowed.details.read().unwrap().key;
                                         let new_score = max_score_merger(
                                             s,
                                             scores.get(&key),
@@ -191,7 +191,7 @@ Expands term with all possible combinations.
 returns All terms that starts with `term` string.
  */
 fn expand_term<I: Debug>(index: &Index<I>, term: &str) -> Vec<String> {
-    let node = find_inverted_index_node(Rc::clone(&index.root), term);
+    let node = find_inverted_index_node(Arc::clone(&index.root), term);
     let mut results = Vec::new();
     if let Some(n) = node {
         expand_term_from_node(n, &mut results, term);
@@ -209,19 +209,19 @@ Recursively goes through inverted index nodes and expands term with all possible
  * `term Term.
  */
 fn expand_term_from_node<I: Debug>(
-    node: Rc<RefCell<InvertedIndexNode<I>>>,
+    node: Arc<RwLock<InvertedIndexNode<I>>>,
     results: &mut Vec<String>,
     term: &str,
 ) {
-    if node.borrow().first_doc.is_some() {
+    if node.read().unwrap().first_doc.is_some() {
         results.push(term.to_owned());
     }
-    let mut child = node.borrow().first_child.clone();
+    let mut child = node.read().unwrap().first_child.clone();
     while let Some(c) = child {
-        let cb = c.borrow();
+        let cb = c.read().unwrap();
         let mut inter = term.to_owned();
         inter.push_str(&String::from(char::from_u32(cb.char_code).unwrap()));
-        expand_term_from_node(Rc::clone(&c), results, &inter); // String.fromCharCode(child.charCode)
+        expand_term_from_node(Arc::clone(&c), results, &inter); // String.fromCharCode(child.charCode)
         child = cb.next.clone();
     }
 }
